/***************************************************************************
 * CVSID: $Id$
 *
 * addon-usb-cable.c : Poll the USB cable for state changes
 *
 * Copyright (C) 2007 Nokia Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 **************************************************************************/

#define _GNU_SOURCE

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <error.h>
#include <poll.h>

#include <glib.h>

#include "libhal/libhal.h"
#include "../../util_helper.h"
#include "../../logger.h"

LibHalContext *ctx;
char *udi;
char *button_type;

static int
emit_mode (char *mode)
{
	DBusError error;

	dbus_error_init (&error);

	libhal_device_set_property_string (ctx, udi,
	                                   "usb_device.mode", mode,
	                                   &error);
	if (dbus_error_is_set (&error)) {
		HAL_ERROR (("Failed to set property: %s", error.message));
		dbus_error_free (&error);
		return 0;
	}

	return 1;
}

static int
emit_state (int state, char *type)
{
	DBusError error;

	dbus_error_init (&error);

	/* XXX: This state file might have to be split to be able to generalize
	 *      this addon*/
	libhal_device_set_property_string (ctx, udi,
	                                   "usb_device.cable_type", type,
	                                   &error);
	if (dbus_error_is_set (&error)) {
		HAL_ERROR (("Failed to set property: %s", error.message));
		dbus_error_free (&error);
		return 0;
	}

	libhal_device_set_property_bool (ctx, udi, "button.state.value", state,
	                                 &error);
	if (dbus_error_is_set (&error)) {
		HAL_ERROR (("Failed to set property: %s", error.message));
		dbus_error_free (&error);
		return 0;
	}

	libhal_device_emit_condition (ctx, udi, "ButtonPressed", button_type,
	                              &error);
	if (dbus_error_is_set (&error)) {
		HAL_ERROR (("Failed to send condition: %s", error.message));
		dbus_error_free (&error);
		return 0;
	}

	return 1;
}

static int
read_mode (int fd, int forreal)
{
	static char old_value[128] = "\0";
	static char sent_value[128] = "\0";
	char value[128];
	int r;

	memset (value, 0, sizeof (value));
	r = read (fd, value, sizeof (value) - 1);
	if (r < 0) {
		HAL_ERROR (("Could not read the value from the file"));
		return r;
	}

	/* Rewind back to beginning, needed by sysfs property files */
	r = lseek (fd, 0, SEEK_SET);
	if (r < 0) {
		HAL_ERROR (("Could not seek in the file"));
		return r;
	}

	if (!forreal)
		return 1;

	/* XXX: The only glib dependency... */
	g_strchomp (value);

	g_debug ("%s: %s (old = %s, sent = %s)", __FUNCTION__,
	         value, old_value, sent_value);

	if (strcmp (old_value, value) == 0) {
		r = 0;
		if (strcmp (sent_value, value) != 0) {
			emit_mode (value);
			strcpy (sent_value, value);
		}
	} else {
		strcpy (old_value, value);
		r = 1;
	}

	return r;
}

static int
read_state (int fd, int forreal)
{
	static char old_value[128] = "\0";
	static char sent_value[128] = "\0";
	char value[128];
	int state;
	int r;

	memset (value, 0, sizeof (value));
	r = read (fd, value, sizeof (value) - 1);
	if (r < 0) {
		HAL_ERROR (("Could not read the value from the file"));
		return r;
	}

	/* Rewind back to beginning, needed by sysfs property files */
	if (lseek (fd, 0, SEEK_SET) < 0) {
		HAL_ERROR (("Could not seek in the file"));
		return r;
	}

	if (!forreal)
		return 1;

	/* XXX: The only glib dependency... */
	g_strchomp (value);

	g_debug ("%s: %s (old = %s, sent = %s)", __FUNCTION__,
	         value, old_value, sent_value);

	if (strcmp (old_value, value) == 0) {
		r = 0;
		state = !!strcmp (value, "disconnected");
		if (strcmp (sent_value, value) != 0) {
			emit_state (state, value);
			strcpy (sent_value, value);
		}
	} else {
		strcpy (old_value, value);
		r = 1;
	}

	return r;
}

struct reader {
	const char *path;
	int flunc;
	int *fd;
	int (*reader)(int fd, int forreal);
};

static int
watch_state (int count, ...)
{
	struct pollfd *pollfd;
	struct reader *readers;
	int r, timeout = 500;
	va_list arg_list;

	pollfd = alloca (sizeof (struct pollfd) * count);
	readers = alloca (sizeof (struct reader) * count);

	va_start (arg_list, count);
	for (r = 0; r < count; r++) {
		readers[r].fd = va_arg (arg_list, int *);
		pollfd[r].fd = *readers[r].fd;
		pollfd[r].events = POLLIN | POLLPRI;
		readers[r].path = va_arg (arg_list, const char *);
		readers[r].flunc = 1;
		readers[r].reader = va_arg (arg_list, int (*)(int fd));
	}
	va_end (arg_list);

	while ((r = poll (pollfd, 2, timeout)) >= 0) {
		int i;

		/* If poll returned due to timeout, reset to indefinite.  */
		g_debug ("Main loop: %d, %d", r, timeout);

		if (r == 0)
			timeout = -1;
		for (i = 0; i < count; i++) {
			g_debug ("%s: %d, %d", readers[i].path,
			         pollfd[i].revents, readers[i].flunc);

			if (r || readers[i].flunc) {
				int rr;

				if (readers[i].path) {
					close (pollfd[i].fd);
					pollfd[i].fd = open (readers[i].path, O_RDONLY);
					*readers[i].fd = pollfd[i].fd;
				}
				rr = readers[i].reader (pollfd[i].fd, !!r ^ readers[i].flunc);
				if (rr < 0)
					return rr;
				if (rr > 0 || pollfd[i].revents) {
					/* If the contents have changed,
					 * rechek in 50 ms. */
					readers[i].flunc = 1;
					timeout = 500;
				} else
					readers[i].flunc = 0;
			}
		}
	}

	return r;
}

int
main (int argc, char **argv)
{
	DBusError error;
	char *sysfs_path, *state_path, *mode_path;
	int fd_state, fd_mode;

	hal_set_proc_title_init (argc, argv);

	setup_logger();

	HAL_INFO (("addon-usb-cable running"));

	udi = getenv("UDI");
	if (!udi) {
		HAL_ERROR (("Failed to get UDI"));
		return 1;
	}

	dbus_error_init (&error);

	ctx = libhal_ctx_init_direct (&error);
	if (ctx == NULL) {
		HAL_ERROR (("Unable to initialise libhal context: %s",
			   error.message));
		return 1;
	}

	if (!libhal_device_addon_is_ready (ctx, udi, &error)) {
		HAL_INFO (("Device %s not ready yet: %s", udi, error.message));
		return 1;
	}

	sysfs_path = getenv ("HAL_PROP_LINUX_SYSFS_PATH");
	if (!sysfs_path) {
		HAL_ERROR (("Property linux.sysfs_path not set for device"));
		return 1;
	}

	HAL_INFO (("sysfs_path=%s", sysfs_path));

	button_type = getenv ("HAL_PROP_BUTTON_TYPE");
	if (!button_type) {
		HAL_ERROR (("Property button.type not set for device"));
		return 1;
	}

	libhal_device_set_property_bool (ctx, udi, "button.has_state", TRUE,
	                                 &error);

	/* XXX: Find a way to not have to hardcode these. */
	if (!asprintf (&state_path, "%s/../cable", sysfs_path)) {
		HAL_ERROR (("Memory allocation problem"));
		return 1;
	}

	if (!asprintf (&mode_path, "%s/../mode", sysfs_path)) {
		HAL_ERROR (("Memory allocation problem"));
		return 1;
	}

	drop_privileges (0);

	hal_set_proc_title ("hald-addon-usb-cable: listening on %s", state_path);

	fd_state = open (state_path, O_RDONLY);
	if (fd_state < 0) {
		HAL_ERROR (("Failed to open '%s'", state_path));
		return 1;
	}

	fd_mode = open (mode_path, O_RDONLY);
	if (fd_mode < 0) {
		HAL_ERROR (("Failed to open '%s'", mode_path));
		return 1;
	}

	read_state (fd_state, 1);
	read_mode (fd_mode, 1);

	watch_state (2, &fd_state, state_path, read_state,
		     &fd_mode, mode_path, read_mode);

	free (state_path);
	free (mode_path);

	close (fd_state);
	close (fd_mode);

	return 0;
}
